/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : pop.cc                                                   */
/*                                                                       */
/*       Date : November  6, 1997                                        */
/* Updated at : November 24, 1997                                        */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include "chromo.h"
#include <stdlib.h>

/*
//
*/

type_Population* definePopulation(int _popSize, unsigned int _chromoLength,
                                  type_Gene *_genePool[])
{

   int i;
   type_Population* myPop = (type_Population*) calloc(1,sizeof(type_Population));

   if(myPop!=NULL)
   {
      
      myPop->myChromoPool = (type_Chromosome **) calloc(_popSize,sizeof(type_Chromosome *));

      if(myPop->myChromoPool!= NULL)
      {
            
          myPop->chromoLength = _chromoLength;
          myPop->genePool = _genePool;
          myPop->popSize = _popSize;

          for (i = 0; i < _popSize; i++)
          {
             myPop->myChromoPool[i] = NULL;
             myPop->myChromoPool[i] = defineChromosome(myPop->chromoLength,myPop->genePool);
             if(myPop->myChromoPool[i]!=NULL) init(myPop->myChromoPool[i],myPop);

          }// end of for

      }
      else
       {
          free(myPop);
          myPop = NULL;
       }
   }
  
   return myPop;

}


type_Population* copyPopulation(type_Population* RHS)
{
   int i;
   type_Population* myPop = definePopulation(RHS->popSize, RHS->chromoLength,
                                             RHS->genePool);

   if(myPop!=NULL)
   {
       myPop->eval = RHS->eval;

       for (i = 0; i < RHS->popSize; i++)
           myPop->myChromoPool[i] = copyChromosome(RHS->myChromoPool[i]);
       
   }// end of if

   return myPop;

}// copyPopulation() 



/*
// Population initialization.
*/

void initPopulation(type_Population* this)
{
   int i;
   this->eval = 0;
   for (i = 0; i < this->popSize; i++) init(this->myChromoPool[i],this);

}// end of init()

/*
// Returns a suitable parent using Roulette Wheel Selection.
*/

type_Chromosome * getParent(type_Population* this)
{
   int i, slump;
   
   slump = rand() % (int)this->eval;

   i = 0;
   while ((i < this->popSize) && (getModEval(this->myChromoPool[i]) < slump)) i++;

   return (this->myChromoPool[i]);

}// end of getParent()

/*
// Sorts chromosome pool in fitness order.
*/

void quickSort(type_Population* this, int start, int finish)
{ 
   int left, right;
   type_Chromosome *tmpChromo;
   eval_t startVal;
   
   left = start;
   right = finish;
   startVal = eval(this->myChromoPool[(start + finish) / 2]);
   do
   {
      while (eval(this->myChromoPool[left]) < startVal)
	 left++;
      while (eval(this->myChromoPool[right]) > startVal)
	 right--;
      if (left <= right)
      {
	 tmpChromo = this->myChromoPool[left];
	 this->myChromoPool[left] = this->myChromoPool[right];
	 this->myChromoPool[right] = tmpChromo;
	 left++;
	 right--;
      }
   }
   while (left < right);
   if (start < right)
      quickSort(this,start, right);
   if (left < finish)
      quickSort(this,left, finish); 
}

/*
//
*/

void replaceWorstChromosomeWith(type_Population* this, type_Chromosome* newChromo)
{
    quickSort(this,0,this->popSize-1);
    this->myChromoPool[this->popSize-1] = copyChromosome(newChromo);

}// end of replaceWorstChromosomeWith()

/*
// Returns the modified fitness of the best cromosome in the population.
*/

eval_long_t fitness(type_Population* this, enum fitness_t fitnessType, 
                    eval_t fitnessParam)
{
   type_Chromosome *temp;
   int popSize = this->popSize;
   int j, i;
   eval_t tmpEval, minEval, maxEval;

   if (this->eval == 0)
   {
      //quickSort(0, popSize - 1);

      for(i=0;i<popSize;i++)
         for(j=0;j<popSize;j++)
	    if(this->myChromoPool[i]!=NULL && this->myChromoPool[j]!=NULL &&
	      (eval(this->myChromoPool[i])<eval(this->myChromoPool[j])) )
	    {
		temp = this->myChromoPool[j];
		this->myChromoPool[j] = this->myChromoPool[i];
		this->myChromoPool[i] = temp;
	    }

      minEval = eval(this->myChromoPool[0]);
      maxEval = eval(this->myChromoPool[popSize - 1]);
      this->eval = 0;

      if (fitnessType == evaluation)
      {  
	 for (i = 0; i < popSize; i++)
	    this->eval = setModEval(this->myChromoPool[i], 
                                    this->eval + maxEval + minEval - eval(this->myChromoPool[i]));
      }
      else if (fitnessType == windowing)
      {
	 eval_t guardedEval;
	 
	 minEval;
	 for (i = 0; i < popSize; i++)
	 {
	    if ((guardedEval = (maxEval - eval(this->myChromoPool[i])))
		< fitnessParam)
	       guardedEval = fitnessParam;
	    this->eval = setModEval(this->myChromoPool[i], this->eval + guardedEval);

	 }// end of for

      }
      
      else // fitnessType == linearNorm
      {
	 this->eval = 0;
	 tmpEval = 100;
	 for (i = 0; i < popSize; i++)
	 {
	    if (tmpEval <= 0)
	    {
	       for (j = i; j < popSize; j++) setModEval(this->myChromoPool[j],this->eval);
	       i = j; // break outer loop
	    }
	    else
	    {
	       this->eval = setModEval(this->myChromoPool[i], this->eval + tmpEval);
	       tmpEval -= fitnessParam;
	    }

	 }// end of for
      }

   }// end of if

   return this->eval;

}// end of fitness()

/*
// Creates a new population using this population as a parent pool
// and returns a pointer to the new population.
// If deletionType == deleteAll, the user of this class has to
// delete this population (the new one is really a _new_ one).
// If deletionType == steadyDelete, the old population is altered
// and thus, requires no further action by the user.
*/

type_Population* newPop(type_Population* this, enum fitness_t fitnessType, 
                        eval_t fitnessParam, float crossover, enum cross_t crossoverType, 
                        float mutation, enum delete_t deletionType, int elitismValue)
{
      int i;      
   int popSize = this->popSize;
   type_Population* retPop = definePopulation(this->popSize,this->chromoLength,
                                              this->genePool);

   type_Chromosome *tmp1 ; 
   type_Chromosome *tmp2 ;
   type_Chromosome *tmpChild ;

   if(retPop!=NULL)
   {
      if (deletionType == deleteAll)
      {
      
        /******************/
        /** copying elite */
        /******************/
        for (i = 0; (i < elitismValue) && (i < popSize); i++)
             retPop->myChromoPool[i] = copyChromosome(this->myChromoPool[i]);

         for (i = elitismValue; i < popSize; i += 2)
	  if (i == (popSize - 1))
	  {
	    tmpChild = copyChromosome(this->myChromoPool[i]);

	    mate(getParent(this), getParent(this),
                 crossover, crossoverType, mutation, retPop->myChromoPool[i], tmpChild);

	    free(tmpChild);
	  }
	  else
	    mate(getParent(this), getParent(this),
                 crossover, crossoverType, mutation, 
                 retPop->myChromoPool[i], retPop->myChromoPool[i+1]);
     }
     else 
     {
       tmp1 = copyChromosome(this->myChromoPool[0]);
       tmp2 = copyChromosome(this->myChromoPool[0]);
      /*********************************/
      /** deletionType == steadyDelete */
      /*********************************/
      
      for (i = 0; i < popSize; i += 2)
      {
         mate(getParent(this), getParent(this), crossover, crossoverType, 
              mutation, tmp1, tmp2);

	 //if (i != (popSize - 1))
	 //*(myChromoPool[popSize - 2]) = *tmp1;
	 if(popSize>1) retPop->myChromoPool[popSize - 2] = copyChromosome(tmp1);

	 //*(myChromoPool[popSize - 1]) = *tmp2;
	 this->myChromoPool[popSize - 1] = copyChromosome(tmp2);
	 
	 this->eval = 0;
	 fitness(this,fitnessType, fitnessParam); // resort
      }
      free(tmp1);
      free(tmp2);
      
     }
     quickSort(retPop,0,retPop->popSize-1);
   }
   
   return retPop;

}// end of newPop()

/*
// Returns the average chromosome fitness of the population.
*/

eval_t fitnessAverage(type_Population* this)
{
   int i;
   eval_long_t sum = 0;

   for (i = 0; i < this->popSize; i++) sum += getEval(this->myChromoPool[i]);

   return (sum / this->popSize);

}// end of fitnessAverage()

eval_t bestFitness(type_Population* this)
{return getEval(this->myChromoPool[0]);}

/* Returns the modified fitness of the best cromosome in the population.  */
eval_t bestFitnessMod(type_Population* this)
{return getModEval(this->myChromoPool[0]);}

/* Returns the average modified chromosome fitness of the population. */
eval_t fitnessAverageMod(type_Population* this)
{return ( this->eval / this->popSize);}

/* Returns the best chromosome in the population. */
type_Chromosome* bestChromo(type_Population* this)
{ return copyChromosome(this->myChromoPool[0]);}

void disposePopulation(type_Population* this)
{
    int i;
    for(i=0;i<this->popSize;i++) free(this->myChromoPool[i]);
    free(this);
}
